/******************************************************************************
*
* Name:     I2C_BF5xx_rev07.asm
*
*******************************************************************************
*
* (C) Copyright 2003 - 2006 - Analog Devices, Inc.  All rights reserved.
*
*Project Name:      I2C programming via PF pins under assembly code
*
*Author:            $Author: ANALOG\tlorenz $
*
*Group:             ADI Systems Engineering Munich
*
*Revision:          $Revision: 0.7 $
*
*Last Modified:     $Date:  12/13/2006$
*
*History:           0.1     01/15/2001  initial release                     TL
*                   0.2     06/06/2003  added support for BF561             GO
*                   0.3     08/25/2004  avoid master/slave driver confliced TL
*                   0.4     10/04/2006  adding error handling               KU
*                                       adding 16bit adddr/data support     KU
*                   0.5     11/01/2006  added derivative independance       TL
*                   0.6     12/04/2006  test run on different EZ-KITs       TL
*                                       "ADVs, Micron, Kodak, Omnivision"   TL
*                   0.7     12/13/2006  added coding style and end test     TL
*
*Software:          VisualDSP++4.5 (November Update)
*
*Hardware:          ADSP-BF561 EZ-KIT LITE REV 2.0 ADSP-BF561 Chip REV 0.5
*                   ADSP-BF533 EZ-KIT LITE REV 1.7 ADSP-BF533 Chip REV 0.4
*
*Connections:       ADDS-HPUSB-ICE
*
*Registers used:    R7-R0, P5-P0
*
*Settings:
*
*Purpose:           I2C Protocol State Machine (SCCB software interface)
*                   This protocol allows to program I2C devices by
*                   Blackfin derivatives via PF pins (GPIOs). It can either be
*                   used for assembly projects or for C projects.
*
*Comments:
*
*Before Calling the SCCB_Interface the inputs to the SCCB_Taskmanager
*must be done.
* Write to the "__SCCB_Control" a "1" for reading from the Device or
* a "2" for writing to the Device.
*
* Leave the total number of bytes in "SCCB_Wordcount".Device Addresses
* and Word Addresses included.
*
* Write all data to the "__SCCB_DataIn".
* e.g.Writing to the device:
*
*         __SCCB_DataIn            Device Address  (LSB must be Zero)
*                                  Word Address
*                                  Data 1
*                                  Data 2
*
*******************************************************************************
* (e.g.Reading from the device:
*         __SCCB_DataIn            Device Address  (LSB must be zero)
*                                  Word Address
*                                  Device Address  (LSB must be one)
*
*
*  The data elements received can be found in the array
*                                  __SCCB_DataOut[]
*
******************************************************************************/

/******************************************************************************
 Include Section
******************************************************************************/
#include "I2C_BF5xx_rev07.h"
/******************************************************************************
 Program Define Section
******************************************************************************/
#define MAX_NUM_ENTRIES     (256)        /* DO NOT set this bigger than 256! */
#define MAX_TRANSACTIONS    (256*2)+ 10  /* max bytes */

/******************************************************************************
 Macros
******************************************************************************/
/* These macros are used in all routine to push/pop all registers */
/* and to reset the timer interrupt again */
#define SCCB_Start_Macro\
        [--SP] = ASTAT;\
        [--SP] = FP;\
        [-- SP] = (R7:0, P5:0);\
        SP+= -64;\
        SP+= -56;\
        link 0x0;\
        [FP+0x8] = R0;\
        [FP+0xC] = R1;\
        [FP + 0x10] = R2

    

#define SCCB_End_Macro\
        p0.h = hi(TIMER_STATUS_REGISTER);\
        p0.l = lo(TIMER_STATUS_REGISTER);\
        R0 = TIMER_INTERRUPT_SERVICE (z);\
        w[p0] = R0;\
        jump.S 2;\
        unlink;\
        SP+=60;\
        SP+=60;\
        (R7:0, P5:0) = [SP++];\
        FP = [SP++];\
        ASTAT = [SP++];\
        SSYNC
/******************************************************************************
 Data
******************************************************************************/
.SECTION L1_data_a;
.ALIGN 4;
.BYTE4        __SCCB_Control;
.BYTE4        __SCCB_Word_Count ;
.BYTE4        __SCCB_DataIn[MAX_TRANSACTIONS];
.BYTE4        __SCCB_DataOut[MAX_TRANSACTIONS];
.BYTE4        __SCCB_In_Progress = 0;
.BYTE4        __SCCB_Error = 0;
.BYTE4        __SCCB_Read_Count = 2;
.BYTE4        SCCB_Point_of_State;
.BYTE4        SCCB_Data_Pointer;
.BYTE4        SCCB_Write_Read_Register;
.BYTE4        SCCB_Bit_Count;
.BYTE4        SCCB_Variable_High_Address;
.BYTE4        SCCB_Read_Start_Cond0_Task_Val;
/*************** SCCB Global Settings ****************************************/
.global     __SCCB_Interface;
.global     __SCCB_Control;
.global     __SCCB_Word_Count;
.global     __SCCB_DataIn;
.global     __SCCB_DataOut;
.global     __SCCB_Read_Count;
.global     __SCCB_In_Progress;
.global     __SCCB_Error;
/******************************************************************************
 Program
******************************************************************************/
.SECTION L1_code;
.ALIGN 4;
__SCCB_Interface:
    SCCB_Start_Macro;
/******************************************************************************
*No register must be kept for this Interface because all contents are
*stored in the variables above. As you can see before entering the
*SCCB_Interface Subroutine all registers will be pushed onto the stack.
*For each access to the SCCB device this subroutine will just be executed
*one time. After execution the pushed registers are poped back again.
*The Timer interrupt initialized in this "asm" file will link to the
*subroutines required for this interface each time it comes up.
******************************************************************************/

/******************************************************************************
 SCCB Timer Interrupt Vector Init 
******************************************************************************/
/* In case the timer will be used for several applications the Interrupt */
/* vetctor for this Interface is set here. The address of the routine to be */
/* executed at first is filled in the interrupt register here. */

    P0.H = hi(ISR_SCCB_Timer);
    P0.L = lo(ISR_SCCB_Timer);  /* Base address of the interrupt vector */
    R0.H =  SCCB_Start_Cond1;
    R0.L =  SCCB_Start_Cond1;
    [P0] = R0;           /* ISR Start Address will be written in the EVT reg */
/******************************************************************************
 SCCB GPIO init as SDA and SCL
******************************************************************************/
    P0.H = hi(System_MMR_High_Address);    /* High address of all the system */
                                           /* registers */
    /* disable PF inputs */
    P0.L = lo(FIO_INEN);
    R0 = w[P0](z);
    bitclr(R0, scl_bit);
    bitclr(R0, sda_bit);
    w[P0] = R0;
    
    /* The Flag PF(SDA) and PF(SCL) shall be high */
    P0.L = lo(FIO_FLAG_S);      
    R0 = SCL (z);
    bitset(R0, sda_bit);
    bitset(R0, scl_bit);
    w[P0] = R0;

    /* Set bit PF(SDA) & PF(SCL) as outputs */
    P0.L = lo(FIO_DIR);
    R0 = w[P0](z);
    bitset(R0, scl_bit);
    bitset(R0, sda_bit);
    w[P0] = R0;
/******************************************************************************
 SCCB Timer Init
******************************************************************************/
    P0.H = hi(System_MMR_High_Address);         /* High address of all the */
                                                /* system registers */
    P0.L = lo(TIMER_STATUS_REGISTER);   
    R0.L = TIMER_INTERRUPT_SERVICE; w[P0] = R0; /* Clear Timer Interrupt */
                                                /* and Overflow bit */
    P0.L = lo(TIMER_DISABLE_REGISTER);          /* disable timer */
    R1 = TIMER_DISABLE_BIT(z);  
    w[P0] = R1;
    
    /* Setup Timerx:  PWM_OUT mode, pulse hi, count to end of           */
    /* period, interrupt, sample TMRx pin, disable pad, EMU_RUN enable  */

    P0.L = lo(TIMER_CONFIG_REGISTER);
    P0.H = hi(TIMER_CONFIG_REGISTER);
    R0.L = TIMER_CONFIG_SET;
    W[P0] = R0.L;

    /* The period is set to a ratio defined in top of this code */
    
    P0.L = lo(TIMER_PERIOD_REGISTER);
    P0.H = hi(TIMER_PERIOD_REGISTER);
    R0 = SCL_PERIOD (z);
    [P0] = R0;

    /* Width provides 50% duty cycle: scl_high = 1/2 of scl_period */

    P0.L = lo(TIMER_WIDTH_REGISTER);
    P0.H = hi(TIMER_WIDTH_REGISTER);
    R0 = SCL_PERIOD >> 1 (z);
    [P0] = R0;
/******************************************************************************
 SCCB Timer Interrupt
******************************************************************************/
    P0.H = hi(System_MMR_High_Address);     /* High address of all the */
                                            /* system registers */
    P0.L = lo(SIC_IMASKx);                  /* pass through timer interrupt */
    R0 = [P0];
    bitset(R0,SIC_IMASKx_BIT);              
    [P0] = R0;

    P0.H = hi(IMASK);
    P0.L = lo(IMASK);
    R0 = [P0];
    bitset(R0,IVEC_SELECT);
    [P0] = R0;                              /* Enable the "IVECx" interrupt */
    
    P0.H = hi(System_MMR_High_Address);     /* High address of all the */
    P0.L = lo(TIMER_ENABLE_REGISTER);       /* system registers enable timer */
    R1 = TIMER_ENABLE_BIT(z);   
    w[P0] = R1;
/******************************************************************************
 SCCB Start Setup
******************************************************************************/
/* prepares the variables used in this Interface */

    P1.H = SCCB_Variable_High_Address;  /* works as a page for all Variables */
    P2.H = SCCB_Variable_High_Address;

    P1.L = SCCB_Bit_Count;
    R0 = 0x8;
    [P1] = R0;

    P1.L = SCCB_Read_Start_Cond0_Task_Val;
    R0 = 0;                             /* due to an extra start condition */
    [P1] = R0;                          /* at reading from device this */
                                        /* variable is required */
    P1.L = __SCCB_DataIn;
    P2.L = SCCB_Data_Pointer;           /* create a moving pointer to */
    [P2] = P1;                          /* the array */
    P2.L = SCCB_Write_Read_Register;
    R0 = [P1];
    [P2] = R0;

    SCCB_End_Macro;
__SCCB_Interface.END:
rts;
/******************************************************************************
 End of Main
*******************************************************************************
 This is the end of the Init program. All other subroutine will be called
 by the Timer Interrupt separately.
******************************************************************************/

/******************************************************************************
 Functions
******************************************************************************/
/******************************************************************************
Each timer interrupt will just call one of the following subroutines.
That restults to a SCCB speed depending of the Timer speed. The time 
between the timer interrupts can be used for processing of user data
there are no loops that would cause a decrease of processor perfomance
******************************************************************************/

/*************** SCCB Start Cond1 ********************************************/
SCCB_Start_Cond1:
    SCCB_Start_Macro;

    P0.H = hi(System_MMR_High_Address); /* High address of all the */
                                        /* system registers */
    P0.L = lo(FIO_FLAG_C);              /* The Flag PF (SDA) will be cleared */
    R0.L = SDA;                         /* at first (Start Condition!) */
    w[P0] = R0;
    
    P0.H = hi(ISR_SCCB_Timer); P0.L = lo(ISR_SCCB_Timer);/* Base address of */
    R0.H =  SCCB_Start_Cond2; R0.L =  SCCB_Start_Cond2; /* the interrupt vec */
    [P0] = R0;                          /* ISR Start Address will be written */
                                        /* in the EVT register */
    SCCB_End_Macro;
rti;

/*************** SCCB Start Cond2 ********************************************/
SCCB_Start_Cond2:

    SCCB_Start_Macro;

    P0.H = hi(System_MMR_High_Address); /* High address of all the system */
                                        /* registers */
    P0.L = lo(FIO_FLAG_C);              /* The Flag PF (SCL) will be cleared */
    R0.L = SCL;                         /* at second(Start Condition!) */
    w[P0] = R0;

    P0.H = hi(ISR_SCCB_Timer); P0.L = lo(ISR_SCCB_Timer);/* Base address */
    R0.H =  SCCB_Send_Data; r0.l =  SCCB_Send_Data;
    [P0] = R0;                          /* ISR Start Address will be written */
                                        /* in the EVT */
    SCCB_End_Macro;
rti;
/*************** SCCB Send Data to the Device*********************************
"SCCB_Send_Data" will send all the addresses and data required for
communication to the device. The data must be shifted through the GPIO Pin
******************************************************************************/

SCCB_Send_Data:

    SCCB_Start_Macro;

    P0.H = hi(System_MMR_High_Address); /* High address of all the system */
    P1.H = SCCB_Variable_High_Address;  /* registers works as a page for all */
                                        /* variables */
    P1.L = SCCB_Write_Read_Register;
    R0 = [P1];

    P1.L = SCCB_Bit_Count;              /* load the bit counter to count the */
    R1 = [P1];
    CC = R1 == 0;
    if CC jump SCCB_Send_DATA_Word_Sent;/* If eight bits are sent get the */
                                        /* next byte */
    CC = bittst(R0, 7);
    if CC jump SCCB_Send_DATA_SET_SDA;  /* Check the data bit in order to */
    P0.L = lo(FIO_FLAG_C);              /* clear or the contents of the bit */
    R5.L = SDA;                         /* check was zero */
    w[P0] = R5;                         /* so the GPIO pin must been cleared */
    jump    SCCB_Bit_Sent_End;
    
SCCB_Send_DATA_SET_SDA: 
    P0.L = lo(FIO_FLAG_S);              /* So, the GPIO pin must been set */
    R5.L = SDA;
    w[P0] = R5;

SCCB_Bit_Sent_End:
    R0 = R0 << 1;                       /* shift the byte to get the next bit */
    P1.L = SCCB_Write_Read_Register;    /* save the shifted byte for the run */
    [P1] = R0;

    P1.L = SCCB_Bit_Count;
    R1 += -1;                           /* decrement the bit counter. */
    [P1] = R1;                          /* So, that last bit can be detected */

    P0.H = hi(ISR_SCCB_Timer); P0.L = lo(ISR_SCCB_Timer); /* Base address */
    R0.H =  SCCB_Write_Clock; R0.L =  SCCB_Write_Clock;
    [P0] = R0;                          /* ISR Start Address will be written */
                                        /* in the EVT */
    SCCB_End_Macro;
rti;

SCCB_Send_DATA_Word_Sent:               /* byte has been sent */
    P0.L = lo(FIO_DIR);     
    R0.L = w[p0];
    bitclr(R0, sda_bit);                /* configure SDA bit as an Input */
    w[P0] = R0;         

    /* enable PF inputs */
    P0.L = lo(FIO_INEN);
    R0 = w[P0](z);
    bitset(R0, sda_bit);
    w[P0] = R0;

    P0.H = hi(ISR_SCCB_Timer); P0.L = lo(ISR_SCCB_Timer);/* Base address */
    R0.H =  SCCB_ACK_TST; R0.L =  SCCB_ACK_TST;
    [P0] = R0;                       /* ISR Start Address will be written in */
                                     /* the EVT register */
    SCCB_End_Macro;
rti;
/*************** SCCB Send Data to the Device finish *************************/

/*************** SCCB SCCB Clock routine**************************************/
SCCB_Write_Clock:

    SCCB_Start_Macro;

    P0.H = hi(System_MMR_High_Address); /* High address of all the system reg */
    P1.H = SCCB_Variable_High_Address;  /* works as a page for all Variables */

    P0.L = lo(FIO_FLAG_S);
    R0.L = W[P0];
    CC = bittst (R0, scl_bit);
    if CC jump SCCB_Set_Write_Clock_Low;/* if scl high, then set it low */
    R0 = SCL(z);                        /* if scl low, then set it high */
    w[P0] = R0;

    SCCB_End_Macro;
rti;     /* remain in this state upon the next interrupt */

SCCB_Set_Write_Clock_Low:
    P0.L = lo(FIO_FLAG_C);
    R0 = SCL(z);
    w[P0] = R0;
    
    P1.L = SCCB_Bit_Count;              /* load the bit counter to count the */
    R1 = [P1];
    CC = R1 == 0;
    if !CC jump JUST_TOGGLE_LOW;        /* If eight bits are sent, */
                                        /* stop driving SDA */
    P0.L = lo(FIO_DIR);     
    R0.L = w[P0];
    bitclr(R0, sda_bit);                /* configure SDA bit as an Input */
    w[P0] = R0;         

    /* enable PF inputs */
    P0.L = lo(FIO_INEN);
    R0 = w[P0](z);
    bitset(R0, sda_bit);
    w[P0] = R0;
    
JUST_TOGGLE_LOW:
    P0.H = hi(ISR_SCCB_Timer); P0.L = lo(ISR_SCCB_Timer);/* Base address */
    R0.H =  SCCB_Send_Data; R0.L =  SCCB_Send_Data;
    [P0] = R0;                          /* ISR Start Address will be written */

    SCCB_End_Macro;
rti;
/*****************************************************************************/
/*************** SCCB SCCB Acknowledge routines*******************************/
SCCB_ACK_Wait:  /* After receiving a word an extra delay is insterted */

    SCCB_Start_Macro;

    P0.H = hi(System_MMR_High_Address);/* High address of all the system reg */
    P1.H = SCCB_Variable_High_Address; /* works as a page for all Variables */
    
    P1.L = __SCCB_Word_Count;
    R1 = [P1];
    CC = R1 == 1;
 /* Read Mode: final word received ends with no acknowledge(SDA driven high) */
    if CC jump No_Acknowledge_By_Master1;  

    P0.H = hi(ISR_SCCB_Timer);
    P0.L = lo(ISR_SCCB_Timer);          /* Int Vector Base Addr */
    R0.H =  SCCB_ACK_TST;
    R0.L =  SCCB_ACK_TST;               /* next state: test for ACK */
    [P0] = R0;
    
    SCCB_End_Macro;
rti;

No_Acknowledge_By_Master1:

    P0.L = lo(FIO_FLAG_S);              /* The Flag PF (SDA) will be set */
    R0.L = SDA;
    w[P0] = R0;

    P0.H = hi(ISR_SCCB_Timer);
    P0.L = lo(ISR_SCCB_Timer);          /* Int Vector Base Addr */
    R0.H =  SCCB_NO_ACK_TST1;
    R0.L =  SCCB_NO_ACK_TST1;           /* next state: toggle SCL */
    [P0] = R0;
    
    SCCB_End_Macro;
rti;

SCCB_NO_ACK_TST1:

    SCCB_Start_Macro;

    P0.H = hi(System_MMR_High_Address);/* High address of all the system reg */
    P1.H = SCCB_Variable_High_Address; /* works as a page for all Variables */
    
    P0.L = lo(FIO_FLAG_S);             /* drive SCL high */
    R0 = SCL(z);
    w[P0] = R0;

    P0.H = hi(ISR_SCCB_Timer);
    P0.L = lo(ISR_SCCB_Timer);         /* Int Vector Base Addr */
    R0.H =  SCCB_NO_ACK_TST2;
    R0.L =  SCCB_NO_ACK_TST2;          /* next state: toggle SCL */
    [P0] = R0;

    SCCB_End_Macro;
rti;

SCCB_NO_ACK_TST2:

    SCCB_Start_Macro;

    P0.H = hi(System_MMR_High_Address);/* High address of all the system reg */
    P1.H = SCCB_Variable_High_Address; /* works as a page for all Variables */

    P0.L = lo(FIO_FLAG_C);             /* drive SCL low */
    R0 = SCL(z);
    w[P0] = R0;

    P0.H = hi(ISR_SCCB_Timer);
    P0.L = lo(ISR_SCCB_Timer);     /* Int Vector Base Addr */
    R0.H =  SCCB_NO_ACK_TST3;
    R0.L =  SCCB_NO_ACK_TST3;      /* next state: release acknowledge on SDA */
    [P0] = R0;

    SCCB_End_Macro;
rti;

SCCB_NO_ACK_TST3:

    SCCB_Start_Macro;

    P0.H = hi(System_MMR_High_Address);/* High address of all the system reg */
    P1.H = SCCB_Variable_High_Address; /* works as a page for all Variables */

    P0.L = lo(FIO_FLAG_C);
    R0 = SDA (z);
    w[P0] = R0;                        /* toggle SCL to low */

    P0.H = hi(ISR_SCCB_Timer);
    P0.L = lo(ISR_SCCB_Timer);         /* Int Vector Base Addr */
    R0.H =  SCCB_Stop_Cond0;
    R0.L =  SCCB_Stop_Cond0;           /* next state: end of transmission */
    [P0] = R0;
    
    SCCB_End_Macro;
rti;


SCCB_ACK_TST:

    SCCB_Start_Macro;

    P0.H = hi(System_MMR_High_Address);/* High address of all the system reg */

    P0.L = lo(FIO_FLAG_S);             /* drive SCL high */
    R0 = SCL(z);
    w[P0] = R0;
    
    P0.L = lo(FIO_FLAG_S);
    R0.L = W[P0];
    CC = bittst (R0,sda_bit);          /* Check SDA input. */
                                       /* It must be low at this point */

    if CC jump SCCB_ACK_Error;         /* If not, something is wrong */

    P0.H = hi(ISR_SCCB_Timer);
    P0.L = lo(ISR_SCCB_Timer);         /* Int Vector Base Addr */
    R0.H =  SCCB_ACK_MNGR;
    R0.L =  SCCB_ACK_MNGR;             /* task manager state */
    [P0] = R0;
    
    SCCB_End_Macro;
rti;
    
SCCB_ACK_MNGR:

    SCCB_Start_Macro;

    P0.H = hi(System_MMR_High_Address);/* High address of all the system reg */
    P1.H = SCCB_Variable_High_Address; /* works as a page for all Variables */

    P0.L = lo(FIO_FLAG_C);
    R0 = SCL (z);
    w[P0] = R0;                        /* toggle SCL to low */

    [--SP] = RETS;                    /* enable nesting; save return address */
    call SCCB_Taskmanger;              /* call the task manager to check job */
    RETS = [sp++];

    SCCB_End_Macro;
rti;

SCCB_ACK_Error:
NOP;
    P0.H = hi(System_MMR_High_Address);/* High address of all the system reg */
    P0.L = lo(FIO_FLAG_S);
    R0.L = W[P0];
    CC = bittst (R0,sda_bit);
    
    if !CC jump SCCB_ACK_Error.glitch;

    P0.H = __SCCB_Error;
    P0.L = __SCCB_Error;
    R0 = 1;
    [P0] = R0;
    jump SCCB_Stop_Cond0_no_push;

SCCB_ACK_Error.glitch:

    P0.H = hi(ISR_SCCB_Timer);
    P0.L = lo(ISR_SCCB_Timer);          /* Int Vector Base Addr */
    R0.H =  SCCB_ACK_MNGR;
    R0.L =  SCCB_ACK_MNGR;              /* acknowledge manager state */
    [P0] = R0;

    SCCB_End_Macro;
rti;
/*****************************************************************************/

/*************** SCCB SCCB Taskmanger routine*********************************
The Taskmanager handles the different use of the interface. It can be 
written to the device or read from the device. Up to ten bytes 
(can be changed) can be written to or read from the device. Finishing the 
data transfer links to the stop condition.
*************** SCCB SCCB Taskmanger routine*********************************/

SCCB_Taskmanger:

    P0.H = hi(System_MMR_High_Address);    /* High address of all the system */
    P1.H = SCCB_Variable_High_Address;     /* works as a page for all Vars */
    P2.H = SCCB_Variable_High_Address;

    P1.L = SCCB_Bit_Count;                 /* The bit counter must be loaded */
    R0 = 0x8;                              /* again after sending each byte */
    [P1] = R0;

    P1.L = __SCCB_Control;
    R0 = [P1];

    /* "1" means read from the device. "2" means write to the device */
    CC = R0 == 2;
    if CC jump SCCB_Write_Task;

    P1.L = __SCCB_Read_Count;
    R1 = [P1];
    CC = R1 == 0;
    if CC jump SCCB_Read1_Task;           /* jump after first 3 bytes got */
    
    P2.L = SCCB_Read_Start_Cond0_Task_Val;
    R2 = [P2];                            /* in read mode after device addr */
    R2 = R1 | R2;                         /* and sub addr has been sent an */
    CC = R2 == 1;                         /* extra start condition must be */
    if CC jump SCCB_Read_Start_Cond0_Task;/* inserted b4 the second device */
                                          /* addr can be sent */
    CC = R1 == -1;
    if CC jump SCCB_Read2_Task;           /* "3" means start reading the */
    R1 += -1;                             /* device otherwise, decrement */
    [P1] = R1;                            /* the count */

SCCB_Write_Task:
    P0.L = lo(FIO_INEN);
    R0.L = w[P0];
    bitclr(R0, sda_bit);                  /* disable SDA input */
    w[P0] = R0;

    P0.L = lo(FIO_DIR);     
    R0.L = w[P0];
    bitset(R0, sda_bit);                  /* configure SDA as output */
    w[P0] = R0;

    P1.L = __SCCB_Word_Count;
    R1 = [P1];
    R1 += -1;
    [P1] = R1;
    CC = R1 == 0;                         /* check the number of bytes sent */
    if CC jump SCCB_Stop_Cond_Task;       /* invoke the stop condition when */
                                          /* all bytes are sent */
    P1.L = SCCB_Data_Pointer;
    P2 = [P1];
    nop;
    nop;
    nop;
    P3 = [P2++];                          /* Increment the pointer to the */
    [P1] = P2;                            /* data array */

    P1.L = SCCB_Write_Read_Register;
    R1 = [P2];
    [P1] = R1;
    
    P0.H = hi(ISR_SCCB_Timer); P0.L = lo(ISR_SCCB_Timer);/* Base address */
    R0.H =  SCCB_Send_Data; R0.L =  SCCB_Send_Data;
    [P0] = R0;  /* ISR Start Address will be written in the EVT register */

RTS;

SCCB_Read_Start_Cond0_Task:
/* insert a delay to place the read start condition */
    P0.H = hi(ISR_SCCB_Timer); P0.L = lo(ISR_SCCB_Timer);/* Base address */
    R0.H =  SCCB_Read_Start_Cond1_Task; R0.L =  SCCB_Read_Start_Cond1_Task;
    [P0] = R0;  /* ISR Start Address will be written in the EVT register */

RTS;

SCCB_Read_Start_Cond1_Task:
    SCCB_Start_Macro;
    
/* In order to generate the read start cond. SDA & SCL must be set to high */
    P0.H = hi(System_MMR_High_Address);/* High address of all the system reg */
    P1.H = SCCB_Variable_High_Address; /* works as a page for all Variables */

    /* disable PF inputs */
    P0.L = lo(FIO_INEN);
    R0 = w[P0](z);
    bitclr(R0, sda_bit);
    w[P0] = R0;
    
    P0.L = lo(FIO_DIR);     
    R0.L = w[P0];
    bitset(R0, sda_bit);                   /* configure SDA as output */
    w[P0] = R0;

    P0.L = lo(FIO_FLAG_S);
    R0 = SCL | SDA(z);
    w[P0] = R0;                            /* set SDA & SCL */

    P0.H = hi(ISR_SCCB_Timer); P0.L = lo(ISR_SCCB_Timer);/* Base address */
    R0.H =  SCCB_Read_Start_Cond2_Task; R0.L =  SCCB_Read_Start_Cond2_Task;
    [P0] = R0;  /* ISR Start Address will be written in the EVT register */

    SCCB_End_Macro;
rti;

SCCB_Read_Start_Cond2_Task:
    SCCB_Start_Macro;
    
/* clear SDA at first to begin the start condition */
    P0.H = hi(System_MMR_High_Address);    /* High address */
    P1.H = SCCB_Variable_High_Address;     /* works as a page */

    P0.L = lo(FIO_FLAG_C);
    R0 = SDA(z);
    w[P0] = R0;                            /* SDA to low */
    
    P1.L = SCCB_Read_Start_Cond0_Task_Val; /* after the read start */
    R0 = 0x8 (z);                          /* it must never apear again */
    [P1] = R0;
    
    P0.H = hi(ISR_SCCB_Timer); P0.L = lo(ISR_SCCB_Timer);/* Base address */
    R0.H =  SCCB_ACK_MNGR; R0.L =  SCCB_ACK_MNGR;
    [P0] = R0;  /* ISR Start Address will be written in the EVT register */

    SCCB_End_Macro;
rti;    

SCCB_Stop_Cond_Task:
    P0.H = hi(ISR_SCCB_Timer); P0.L = lo(ISR_SCCB_Timer);/* Base address */
    R0.H =  SCCB_Stop_Cond0; R0.L =  SCCB_Stop_Cond0;
    [P0] = R0;  /* ISR Start Address will be written in the EVT register */

RTS;

SCCB_Read1_Task:
    P1.L = __SCCB_Read_Count;
    R0 = -1;
    [P1] = R0;

    P1.L = SCCB_Write_Read_Register;       /* Get the data to send */
    R0 = 0;
    [P1] = R0;

    P1.L = __SCCB_DataOut;
    P2.L = SCCB_Data_Pointer;
    [P2] = P1;

SCCB_Read2_Task:

    P1.L = __SCCB_Word_Count;
    R1 = [P1];
    R1 += -1;
    [P1] = R1;

                                        /* make sure SDA is configured as in */
    P0.L = lo(FIO_DIR);
    R0 = w[P0](z);
    bitclr(R0, sda_bit);
    w[P0] = R0;

                                        /* enable sda input */
    P0.L = lo(FIO_INEN);
    R0 = w[P0](z);
    bitset(R0, sda_bit);
    w[P0] = R0;
        
    P0.H = hi(ISR_SCCB_Timer); P0.L = lo(ISR_SCCB_Timer);   /* Base address */
    R0.H =  SCCB_Rcv_Clk_Test; R0.L =  SCCB_Rcv_Clk_Test;
    [P0] = R0;  /* ISR Start Address will be written in the EVT register */

RTS;
/******************** SCCB Receive Data***************************************/
SCCB_Receive_Data:

    SCCB_Start_Macro;

    P0.H = hi(System_MMR_High_Address);/* High address */
    P1.H = SCCB_Variable_High_Address; /* works as a page */
    P2.H = SCCB_Variable_High_Address;

    P1.L = SCCB_Write_Read_Register;   /* Get the current rcv byte */
    R0 = [P1];
    R0 = R0 << 1;                      /* shift the byte to get the next bit */

    P1.L = __SCCB_Word_Count;
    R1 = [P1];
    CC = R1 == 0;                      /* check the number of bytes already */
    if CC jump SCCB_Stop_Receive_Data; /* received */

    P0.L = lo(FIO_FLAG_S);
    R5.L = w[p0];
    CC = bittst (R5,sda_bit);          /* Check the SDA bit. */
    if CC jump SCCB_Receive_High_detected;  
    
    BITCLR (R0,0);
    jump    SCCB_End_of_Receive_Bit;

SCCB_Receive_High_detected:
    bitset (R0,0);

SCCB_End_of_Receive_Bit:

    P0.L = lo(FIO_FLAG_C);
    R5.L = SCL;                         /* drive SCL low */
    w[P0] = R5;

    P1.L = SCCB_Write_Read_Register;    /* save the shifted byte for the run */
    [P1] = R0;

    P1.L = SCCB_Bit_Count;              /* update the bit count */
    R1 = [P1];
    R1 += -1;
    [P1] = R1;
    CC = R1 == 0;                       /* check if it was the last bit */
    if CC jump SCCB_Receive_Data_Word_Sent;/* jump when last bit */

    P0.H = hi(ISR_SCCB_Timer);
    P0.L = lo(ISR_SCCB_Timer);          /* Base address */
    R0.H =  SCCB_Rcv_Clk_Test;
    R0.L =  SCCB_Rcv_Clk_Test;          /* next state: run clock cycle */
    [P0] = R0;

    SCCB_End_Macro;
rti;

SCCB_Receive_Data_Word_Sent:

    P1.L = SCCB_Data_Pointer;
    P2 = [P1];
    P3.L = SCCB_Write_Read_Register;    /* save the shifted byte for the run */
    P3.H = SCCB_Write_Read_Register;
    R0 = [P3];
    [P2] = R0;
    R0 = 0;
    [P3] = R0;                          /* clear the content of */
    nop;                                /* "SCCB_Write_Read_Register" */
    nop;                                /* for next byte */
    nop;
    P3 = [P2++];                        /* Increment the pointer to the data */
    [P1] = P2;                          /* array */

                                        /* acknowledge the data received */
                                        /* disable sda input */
    P0.L = lo(FIO_INEN);
    R0 = w[P0](z);
    bitclr(R0, sda_bit);
    w[P0] = R0;

    P0.L = lo(FIO_DIR);     
    R0.L = w[P0];
    bitset(R0, sda_bit);                /* configure SDA as output */
    w[P0] = R0;
    
    P1.L = __SCCB_Word_Count;
    R1 = [P1];
    CC = R1 == 1;
    if CC jump No_Acknowledge_By_Master0;  

                                        /* pull low the SDA line */
    P0.H = hi(System_MMR_High_Address); /* High address of all the system reg */
    P0.L = lo(FIO_FLAG_C);              /* The Flag PF (SDA) will be cleared */
    R0.L = SDA;                         /* at first (Start Condition!) */
    w[P0] = R0;

    P0.H = hi(ISR_SCCB_Timer);
    P0.L = lo(ISR_SCCB_Timer);          /* Base address of the interrupt vector */

    R0.H = SCCB_ACK_Wait;
    R0.L = SCCB_ACK_Wait;
    [P0] = R0;
    
    
    SCCB_End_Macro;
rti;

No_Acknowledge_By_Master0:  

    P0.H = hi(ISR_SCCB_Timer);
    P0.L = lo(ISR_SCCB_Timer);          /* Int Vector Base Addr */
    
    R0.H =  SCCB_ACK_Wait;
    R0.L =  SCCB_ACK_Wait;
    [P0] = R0;

    SCCB_End_Macro;
rti;

SCCB_Stop_Receive_Data:
    /* disable sda input */
    P0.L = lo(FIO_INEN);
    R0 = w[P0](z);
    bitclr(R0, sda_bit);
    w[P0] = R0;
    
    P0.L = lo(FIO_DIR);     
    R0.L = w[P0];
    bitset(R0, sda_bit);                /* configure SDA as output  */
    w[P0] = R0;

    P0.H = hi(ISR_SCCB_Timer); P0.L = lo(ISR_SCCB_Timer);/* Base address */
    R0.H =  SCCB_Stop_Cond2; R0.L =  SCCB_Stop_Cond2;
    [P0] = R0;  /* ISR Start Address will be written in the EVT register */

    SCCB_End_Macro;
rti;
/*************** SCCB Receive Clock Test routine******************************/

SCCB_Rcv_Clk_Test:

    SCCB_Start_Macro;

    P0.H = hi(System_MMR_High_Address); /* High address */
    P1.H = SCCB_Variable_High_Address;  /* works as a page for all Variables */

    P0.L = lo(FIO_FLAG_S);
    R0.L = W[P0];
    CC = bittst (R0, scl_bit);
    if CC jump SCCB_Set_Rcv_Clk_Lo;     /* if scl high, jump to clear it */
    
    /* if scl low, then set it high in the next state */
    P0.H = hi(ISR_SCCB_Timer);
    P0.L = lo(ISR_SCCB_Timer);          /* interrupt vector base addr */
    R0.H =  SCCB_Rcv_Clk_Hi;
    R0.L =  SCCB_Rcv_Clk_Hi;            /* next state: run clock cycle */
    [P0] = R0;
    
    SCCB_End_Macro;
rti;

SCCB_Set_Rcv_Clk_Lo:

    P0.L = lo(FIO_FLAG_C);
    R0 = SCL(z);
    w[P0] = R0;                         /* set SCL low */

    SCCB_End_Macro;
rti;      /* stay in this state until the next interrupt */

/*****************************************************************************/
/*************** SCCB Receive Clock Hi routine********************************/
SCCB_Rcv_Clk_Hi:

    SCCB_Start_Macro;

    P0.H = hi(System_MMR_High_Address);     /* High address */
    P0.L = lo(FIO_FLAG_S);
    R0 = SCL (z);
    w[P0] = R0;                             /* set SCL high */
    
    P0.H = hi(ISR_SCCB_Timer);
    P0.L = lo(ISR_SCCB_Timer);              /* interrupt vector base addr */
    R0.H =  SCCB_Receive_Data;              /* next state: receive data */
    R0.L =  SCCB_Receive_Data;              /* from device */
    [P0] = R0;
    
    SCCB_End_Macro;
rti;
/*****************************************************************************/

/*************** SCCB Stop Cond0,1 *******************************************/
/* Invoke stop condition if the transmission is complete */
SCCB_Stop_Cond0:                /* do nothing (just a delay) */

    SCCB_Start_Macro;
SCCB_Stop_Cond0_no_push:
    P0.H = hi(ISR_SCCB_Timer); P0.L = lo(ISR_SCCB_Timer);/* Base address */
    R0.H =  SCCB_Stop_Cond1; R0.L =  SCCB_Stop_Cond1;
    [P0] = R0;               /* ISR Start Address will be written in the EVT */

    SCCB_End_Macro;
rti;

SCCB_Stop_Cond1:

    SCCB_Start_Macro;

    P0.H = hi(System_MMR_High_Address); /* High address */
    P0.L = lo(FIO_FLAG_S);  
    R0 = SCL(z);
    w[P0] = R0;

    P0.H = hi(ISR_SCCB_Timer); P0.L = lo(ISR_SCCB_Timer);/* Base address */
    R0.H =  SCCB_Stop_Cond2; R0.L =  SCCB_Stop_Cond2;
    [P0] = R0;  /* ISR Start Address will be written in the EVT register */

    SCCB_End_Macro;
rti;
/*************** SCCB Stop Cond2 *********************************************/
SCCB_Stop_Cond2:

    SCCB_Start_Macro;

    P0.H = hi(System_MMR_High_Address);     /* High address */
    P0.L = lo(FIO_FLAG_S);                  /* set SDA high to stop */
    R0 = SDA (z);
    w[P0] = R0;

    P0.H = hi(ISR_SCCB_Timer); P0.L = lo(ISR_SCCB_Timer);/* Base address */
    R0.H =  SCCB_End_of_Transmission; R0.L =  SCCB_End_of_Transmission;
    [P0] = R0;  /* ISR Start Address will be written in the EVT register */

    SCCB_End_Macro;
rti;
/*************** SCCB Stop Cond2 *********************************************/
/* The transmission ends and all the timer settings must be disabled */
/* This would be a good place to set a bit to replay the transfer is done */
SCCB_End_of_Transmission:

    SCCB_Start_Macro;
    
    P0.H = hi(System_MMR_High_Address);     /* High address */
    P0.L = lo(TIMER_DISABLE_REGISTER);      /* disable timer */
    R1 = TIMER_DISABLE_BIT(z);  
    w[P0] = R1; /* disable the timer */
    
    P0.H = hi(System_MMR_High_Address);     /* High address */
    P0.L = lo(TIMER_STATUS_REGISTER);   
    R0 = TIMER_INTERRUPT_SERVICE_END(z);
    w[P0] = R0; /* disable the timer clear interrupt bit and overflow bit */

    P0.L = lo(SIC_IMASKx);                  /* disable mask the Timerx again */
    R0 = [P0];
    bitclr(R0,SIC_IMASKx_BIT);
    [P0] = R0;
    
    P0.H = hi(IMASK);
    P0.L = lo(IMASK);
    R0 = [P0];
    bitclr(R0,IVEC_SELECT);
    [P0] = R0;
    
    P0.H = __SCCB_In_Progress;               /* keeps the core away from */
    P0.L = __SCCB_In_Progress;               /* calling the SCCB twice */
    R0 = 0;
    [P0] = R0;
    
    SCCB_End_Macro;

rti;
/******************************************************************************
 End of Functions
******************************************************************************/

/******************************************************************************
 End of Program
******************************************************************************/
